#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#    Bass Diagrams, draw and export simple fretboard diagrams for bass players
#
#    Copyright (C) 2020  Marco De Freitas
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations
import tempfile
import tkinter
from tkinter import filedialog
from abc import ABC
import argparse

from PIL import Image

FACTOR=4

X_ORIGIN = 60
Y_ORIGIN = 60
STRING_SPACING = 27 * FACTOR
STRING_LENGTH = 800 * FACTOR
FRET_SPACING = 50 * FACTOR
FRET_WIDTH = 1 # * FACTOR
DOTS_DIAMETER = 10 * FACTOR
INTERACTIV_DOTS_DIAMETER = 24 * FACTOR
NB_OF_STRINGS = 4
STRINGS_WIDTH = 1 * FACTOR
SILLET_WIDTH = 4 * FACTOR
OFFSET = 10
DOT_POSITION = [3, 5, 7, 9, 15]


class Mediator(ABC):
    """
    The Mediator interface declares a method used by components to notify the
    mediator about various events. The Mediator may react to these events and
    pass the execution to other components.
    """

    def notify(self, sender: object, event: str) -> None:
        pass


class ConcreteMediator(Mediator):
    def __init__(self, canvas: MyCanvas, buttons: ButtonsBar, master: tkinter.Tk) -> None:
        super(ConcreteMediator, self).__init__()
        self._canvas = canvas
        self._canvas.mediator = self
        self._buttons = buttons
        self._buttons.mediator = self
        self._master = master

    def notify(self, sender: object, event: str, **kwargs) -> None:
        if event == "4":
            self._canvas.set_string_nb(4)

        elif event == "5":
            self._canvas.set_string_nb(5)

        elif event == "6":
            self._canvas.set_string_nb(6)

        elif event == "erase":
            self._canvas.erase()

        elif event == "rotate":
            self._canvas.rotate(self._master)

        elif event == "export":
            self._canvas.export()


class BaseComponent:
    """
    The Base Component provides the basic functionality of storing a mediator's
    instance inside component objects.
    """

    def __init__(self, mediator: Mediator = None) -> None:
        self._mediator = mediator

    @property
    def mediator(self) -> Mediator:
        return self._mediator

    @mediator.setter
    def mediator(self, mediator: Mediator) -> None:
        self._mediator = mediator


class ButtonsBar(tkinter.Frame, BaseComponent):
    """
    Displays buttons for user action
    """

    def __init__(self):
        super(ButtonsBar, self).__init__()

        button1 = tkinter.Button(self,
                                 text="4 cordes",
                                 command=lambda: self.mediator.notify(self, "4"),
                                 anchor=tkinter.CENTER,
                                 highlightthickness=0)
        button1.configure(width=5,
                          background="white",
                          activebackground="#33B5E5",
                          relief=tkinter.FLAT)
        button1.pack(side='left')
        button2 = tkinter.Button(self,
                                 text="5 cordes",
                                 command=lambda: self.mediator.notify(self, "5"),
                                 anchor=tkinter.CENTER,
                                 highlightthickness=0)
        button2.configure(width=5,
                          background="white",
                          activebackground="#33B5E5",
                          relief=tkinter.FLAT)
        button2.pack(side='left')
        button3 = tkinter.Button(self,
                                 text="6 cordes",
                                 command=lambda: self.mediator.notify(self, "6"),
                                 anchor=tkinter.CENTER,
                                 highlightthickness=0)
        button3.configure(width=5,
                          background="white",
                          activebackground="#33B5E5",
                          relief=tkinter.FLAT)
        button3.pack(side='left')
        button4 = tkinter.Button(self,
                                 text="orientation",
                                 command=lambda: self.mediator.notify(self, "rotate"),
                                 anchor=tkinter.CENTER,
                                 highlightthickness=0)
        button4.configure(width=10,
                          background="white",
                          activebackground="#33B5E5",
                          relief=tkinter.FLAT)
        button4.pack(side='left')
        button5 = tkinter.Button(self,
                                 text="effacer",
                                 command=lambda: self.mediator.notify(self, "erase"),
                                 anchor=tkinter.CENTER,
                                 highlightthickness=0)
        button5.configure(width=10,
                          background="white",
                          activebackground="#33B5E5",
                          relief=tkinter.FLAT)
        button5.pack(side='left')
        button6 = tkinter.Button(self,
                                 text="exporter",
                                 command=lambda: self.mediator.notify(self, "export"),
                                 anchor=tkinter.CENTER,
                                 highlightthickness=0)
        button6.configure(width=10,
                          background="white",
                          activebackground="#33B5E5",
                          relief=tkinter.FLAT)
        button6.pack(side='left')

class MyEntry(tkinter.Entry):
    def __init__(self, *args, **kwargs):
        super(MyEntry, self).__init__(*args, **kwargs)

    def changeBackground(self, event):
        if self.cget('bg')== 'white':
            self['bg']="Grey"
        elif self.cget('bg')=='Grey':
            self['bg']= 'white'

class MyCanvas(tkinter.Canvas):
    """
    Canvas, background draws bass fretboard with strings, frets and dots. On top off it interactive dots.
    """

    def __init__(self, *args, **kwargs):
        super(MyCanvas, self).__init__(*args, **kwargs)
        # self.buttons_frame = ButtonsBar()
        # self.buttons_frame.pack(side='top')
        self.tk_focusFollowsMouse()
        self.state = 'Landscape'
        self._background(x_origin=X_ORIGIN,
                         y_origin=Y_ORIGIN,
                         string_spacing=STRING_SPACING,
                         string_length=STRING_LENGTH,
                         frets_spacing=FRET_SPACING,
                         dots_diameter=DOTS_DIAMETER,
                         interactiv_dots_diameter=INTERACTIV_DOTS_DIAMETER,
                         nb_of_strings=4)
        self.pack(fill='both', expand=True)
        self.bind("<Button-1>", self.click)
        self.bind("<Button-3>", self.click_droit)
        # self.bind("<Double-Button-3>", self.double_click_droit)

    def _draw_strings(self, nb_of_strings, x_origin, string_length, y_origin, string_spacing):
        x_end = x_origin + string_length
        for j in range(nb_of_strings):
            self.create_line(x_origin, y_origin + j * string_spacing,
                             x_end, y_origin + j * string_spacing,
                             width=1)

    def _draw_sillet_and_frets(self, x_origin, y_origin, nb_of_strings, string_spacing, frets_spacing):
        y_end = y_origin + (nb_of_strings - 1) * string_spacing
        # Draw sillet
        self.create_line(x_origin + 1, y_origin + 1,
                         x_origin + 1, y_end,
                         width=4, capstyle='round')
        # Draw frets
        for i in range(17):
            self.create_line(x_origin + i * frets_spacing, y_origin,
                             x_origin + i * frets_spacing, y_end,
                             width=FRET_WIDTH, capstyle='butt')

    def _draw_simple_dots(self, x_origin, y_origin, frets_spacing, string_spacing, dotsize, nb_of_strings, remplissage):
        dot_y_top = (y_origin + string_spacing * (nb_of_strings - 1) / 2) - (dotsize/2)
        dot_x_origin = - x_origin + (dotsize/2) - (frets_spacing/2)
        for i in DOT_POSITION:
            i = i-1
            self.create_oval(i * frets_spacing - dot_x_origin, dot_y_top,
                             i * frets_spacing - dot_x_origin + dotsize, dot_y_top + dotsize,
                             fill=remplissage)

    def _draw_double_dots(self, x_origin, y_origin, dotsize, nb_of_strings, string_spacing, frets_spacing, remplissage):
        dot_x_origin = - x_origin + (dotsize / 2) - (frets_spacing / 2)
        double_dot_1_y_top = y_origin + dotsize + (nb_of_strings - 4) * string_spacing / 2
        double_dot_1_y_bottom = y_origin + 2 * dotsize + (nb_of_strings - 4) * string_spacing / 2
        double_dot_2_y_top = y_origin + dotsize + 2 * string_spacing + (nb_of_strings - 4) * string_spacing / 2
        double_dot_2_y_bottom = y_origin + 2 * dotsize + 2 * string_spacing + (nb_of_strings - 4) * string_spacing / 2
        # double dot
        self.create_oval(11 * frets_spacing - dot_x_origin,
                         double_dot_1_y_top,
                         11 * frets_spacing - dot_x_origin + dotsize,
                         double_dot_1_y_bottom,
                         fill=remplissage)
        self.create_oval(11 * frets_spacing - dot_x_origin,
                         double_dot_2_y_top,
                         11 * frets_spacing - dot_x_origin + dotsize,
                         double_dot_2_y_bottom,
                         fill=remplissage)

    def _draw_interactiv_dots(self, x_origin, y_origin, string_spacing, nb_of_strings, frets_spacing,
                              interactiv_dots_diameter, offset):
        dots_list = []
        radius = interactiv_dots_diameter / 2


        for i in range(nb_of_strings):
            dots_list.append(self.create_oval(x_origin - radius,
                                              y_origin - radius + i * string_spacing,
                                              x_origin - radius + interactiv_dots_diameter,
                                              y_origin - radius + i * string_spacing + interactiv_dots_diameter,
                                              tags='activ_dot', outline='', activefill="blue"))
            for j in range(17):
                j += 1
                dots_list.append(self.create_oval(j * frets_spacing  + x_origin - radius,
                                                  y_origin - radius + (string_spacing * i),
                                                  j * frets_spacing  + interactiv_dots_diameter + x_origin - radius,
                                                  y_origin + radius + (string_spacing * i),
                                                  tags='activ_dot', outline='', activefill="blue"))

        print(dots_list)

    # def _entry_creator(self):
    #     entree = MyEntry(self, bd=0, width=10, highlightthickness=0, justif=fdssa)
    #     return entree

    # FIXME change open string interactiv dots to standard dot shape, because no transparency issues with de entry-window
    def _background(self, x_origin, y_origin,
                    string_spacing, string_length,
                    frets_spacing, dots_diameter, interactiv_dots_diameter,
                    nb_of_strings):

        self.nbString = nb_of_strings

        self._draw_strings(nb_of_strings, x_origin, string_length, y_origin, string_spacing)
        self._draw_sillet_and_frets(x_origin, y_origin, nb_of_strings, string_spacing, frets_spacing)
        self._draw_simple_dots(x_origin,y_origin, frets_spacing, string_spacing, dots_diameter, nb_of_strings, remplissage='')
        self._draw_double_dots(x_origin, y_origin, dots_diameter, nb_of_strings, string_spacing, frets_spacing, remplissage='')
        self._draw_interactiv_dots(x_origin, y_origin, string_spacing, nb_of_strings, frets_spacing,
                                   interactiv_dots_diameter, OFFSET)
        # self.entry_list = [self._entry_creator() for i in range(17 * nb_of_strings)]

    def set_string_nb(self, nb):
        """ backgroud is redrawed accordingly to user action on buttons (4-5-6 strings) and Portrait/Landscape state.
        """
        self.delete(tkinter.ALL)
        self._background(x_origin=X_ORIGIN,
                         y_origin=Y_ORIGIN,
                         string_spacing=STRING_SPACING,
                         string_length=STRING_LENGTH,
                         frets_spacing=FRET_SPACING,
                         dots_diameter=DOTS_DIAMETER,
                         interactiv_dots_diameter=INTERACTIV_DOTS_DIAMETER,
                         nb_of_strings=nb)
        if self.state == "Portrait":
            self.state = "Landscape"
            self.mediator.notify(self, "rotate", geom=True)


    def _reverse_status(self, item):
       # print(self.itemconfig(item))
        if self.itemcget(item, 'fill') == '':
            self.itemconfig(item, fill="grey", outline='grey')
            self.update_idletasks()
        else:
            self.itemconfig(item, fill="", outline='')

    def click(self, event):
        """ User click change color of interactive dots"""
        # FIXED create a function called reverse_status to turn dot fill Grey/white
        if self.find_withtag(tkinter.CURRENT):
            if self.itemcget(tkinter.CURRENT, 'tags') == 'activ_dot current':
                self._reverse_status(tkinter.CURRENT)
                current_box = self.bbox(tkinter.CURRENT)
                print(current_box)
                x0  = current_box [0]
                y0 = current_box [1]
                x1 = current_box [2]
                y1 = current_box [2]
                enclosed_entry = self.find_enclosed(x0, y0, x1, y1)
                for item in enclosed_entry:
                    if item > 100:
                        print(self.itemconfig(item)) # Fixme we catch the window item, not the Entry within!!!
                        above = self.find_above(item)
                        if above:
                            print(self.itemconfig(above))







    # entry on right click create entry with background color similar to activated interactive dot

    def click_droit(self, event):
        # if self.itemcget(tkinter.CURRENT, 'tags') == 'entry':
        #     entry = self.itemcget(tkinter.CURRENT)
        #     canvas.delete(entry)
        print(self.itemconfig(tkinter.CURRENT))
        if self.itemcget(tkinter.CURRENT, 'fill') == 'grey':
            background = 'grey'
        else:
            background = 'white'
        if self.itemcget(tkinter.CURRENT, 'tags') != 'entry':
            coords = self.coords(self.find_withtag(tkinter.CURRENT))
            entree = MyEntry(self, bd=0, width=4, highlightthickness=0, bg=background, font=('Helvetica', '25'),
                                   justify='center', validate='focus')
            entree.bind_class('Entry','<Button-3>', entree.changeBackground)
            self.create_window(coords[0] + 13 * FACTOR, coords[1] + 12 * FACTOR, window=entree, tags='entry')
        else:
            print("boom")


    # def double_click_droit(self, event):
    #     if self.itemcget(tkinter.CURRENT, 'tags') == 'entry':
    #         entry = self.find_withtag(tkinter.CURRENT)
    #         print(entry)
    #         canvas.delete(entry)

    # else:
    #    print("prout")

    def erase(self):
        """Change all interactive dots fill back to transparent"""
        dots = self.find_withtag('activ_dot')
        for item in dots:
            self.itemconfigure(item, outline='', fill='')
        entries = self.find_withtag('entry')
        for item in entries:
            self.delete(item)

    def export(self):
        """Export canvas as png"""
        filename = tkinter.filedialog.asksaveasfilename(defaultextension='.png',
                                                        filetypes=[("PNG", "*.png"), ("EPS", "*.eps")])
        print(filename)
        if not filename:
            return
        else:
            temporary_file = tempfile.NamedTemporaryFile()
            self.postscript(colormode='gray', file=temporary_file.name, rotate=False, x=1, y=1)
            with Image.open(temporary_file) as image:
                image.save(filename)

    def rotate(self, master: tkinter.Tk, geom=False):
        """change background orientation Portrait/Landscape"""
        geometry_string = master.geometry()
        # print(geometry_string)
        x = geometry_string[:3]
        y = geometry_string[4:7]
        new_geometry = ''.join((y, 'x', x))
        # FIXME why calculate this at runtime, why not have constant values
        if self.nbString == 4:
            factor = 160 * FACTOR
        if self.nbString == 5:
            factor = 187 * FACTOR
        if self.nbString == 6:
            factor = 214 * FACTOR
        # FIXME very ugly nested ifs
        if self.state == 'Landscape':
            for item in self.find_all():
                if len(self.coords(item)) == 4:
                    x1, y1, x2, y2 = self.coords(item)
                    self.coords(item, factor - y1, x1, factor - y2, x2)
                else:
                    x1, y1 = self.coords(item)
                    self.coords(item, factor - y1, x1)
                self.state = 'Portrait'

        else:
            for item in self.find_all():
                if len(self.coords(item)) == 4:
                    x1, y1, x2, y2 = self.coords(item)
                    self.coords(item, y1, factor - x1, y2, factor - x2)
                else:
                    x1, y1 = self.coords(item)
                    self.coords(item, y1, factor - x1)

                self.state = 'Landscape'

        if geom is False:
            master.geometry(new_geometry)

    def activate_dot(self, id):
        """For command line interface, change fill of interactiv_dot"""
        self._reverse_status(id)

# Init GUI
def create_master():
    """Create and configure master window."""
    master_window = tkinter.Tk()
    master_window.configure(background="white")
    master_window.resizable(True, True)
    master_window.title('Diagrams')
    return master_window


def create_buttons():
    """Create and configure buttons bar."""
    buttons_bar = ButtonsBar()
    buttons_bar.pack(side='top')
    return buttons_bar

# command line utilities
def parse_dots_input(dots_input):
    string_val=[[],[],[],[],[],[]]
    i = 0
    for item in dots_input:
        if item == "/":
            i += 1
        elif item != "n":
            string_val[i].append(int(item))
    return string_val

def convert_frets_to_ids(init_id=int, frets_per_string=int, string_val=list):
    activ_dots = []
    id0=init_id
    for i in range(len(string_val)):
        if len(string_val[i]) == 0 :
            pass
        else:
            for j in string_val[i]:
                activ_dots.append(j+id0)
        id0 = init_id +  (i+1) * 18
    return activ_dots

# Start GUI
def start_gui(*args):
    master = create_master()

    buttons = create_buttons()
    canvas = MyCanvas(master, width=800, height=500, background='#ffffff', closeenough=1, highlightthickness=0)
    mediator = ConcreteMediator(canvas, buttons, master)
    if len(args) == 1:
        for item in args[0]:
            canvas._reverse_status(item)
    master.mainloop()



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Diagrams.py command line interface")
    parser.add_argument("dots", type=str, action="extend", nargs="*",
                        help="interactive dots positions (space separator) for each string (slash separator). example C major scale: 2 4 5 / 2 3 5 / 3 5")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="increase output verbosity")
    args = parser.parse_args()
    dots = args.dots
    if not dots:
        start_gui()
    else:
        parsed_dots = parse_dots_input(dots)
        dots_list = convert_frets_to_ids(30, 17, parsed_dots)
        start_gui(dots_list)



